@@ -13,13 +13,26 @@
 # limitations under the License.
 
 """Bits and pieces used by the driver that don't really fit elsewhere."""
+from __future__ import annotations
 
 import sys
 import traceback
 from collections import abc
-from typing import Any, List, NoReturn
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Container,
+    Iterable,
+    Mapping,
+    NoReturn,
+    Optional,
+    Sequence,
+    TypeVar,
+    Union,
+    cast,
+)
 
-from bson.son import SON
 from pymongo import ASCENDING
 from pymongo.errors import (
     CursorNotFound,
@@ -34,8 +47,13 @@ from pymongo.errors import (
 )
 from pymongo.hello import HelloCompat
 
+if TYPE_CHECKING:
+    from pymongo.cursor import _Hint
+    from pymongo.operations import _IndexList
+    from pymongo.typings import _DocumentOut
+
 # From the SDAM spec, the "node is shutting down" codes.
-_SHUTDOWN_CODES = frozenset(
+_SHUTDOWN_CODES: frozenset = frozenset(
     [
         11600,  # InterruptedAtShutdown
         91,  # ShutdownInProgress
@@ -65,19 +83,25 @@ _RETRYABLE_ERROR_CODES: frozenset = _NOT_PRIMARY_CODES | frozenset(
         89,  # NetworkTimeout
         9001,  # SocketException
         262,  # ExceededTimeLimit
+        134,  # ReadConcernMajorityNotAvailableYet
     ]
 )
 
 # Server code raised when re-authentication is required
-_REAUTHENTICATION_REQUIRED_CODE = 391
+_REAUTHENTICATION_REQUIRED_CODE: int = 391
+
+# Server code raised when authentication fails.
+_AUTHENTICATION_FAILURE_CODE: int = 18
 
 
-def _gen_index_name(keys):
+def _gen_index_name(keys: _IndexList) -> str:
     """Generate an index name from the set of fields it is over."""
     return "_".join(["{}_{}".format(*item) for item in keys])
 
 
-def _index_list(key_or_list, direction=None):
+def _index_list(
+    key_or_list: _Hint, direction: Optional[Union[int, str]] = None
+) -> Sequence[tuple[str, Union[int, str, Mapping[str, Any]]]]:
     """Helper to generate a list of (key, direction) pairs.
 
     Takes such a list, or a single key, or a single key and direction.
@@ -89,58 +113,73 @@ def _index_list(key_or_list, direction=None):
     else:
         if isinstance(key_or_list, str):
             return [(key_or_list, ASCENDING)]
-        if isinstance(key_or_list, abc.ItemsView):
-            return list(key_or_list)
+        elif isinstance(key_or_list, abc.ItemsView):
+            return list(key_or_list)  # type: ignore[arg-type]
+        elif isinstance(key_or_list, abc.Mapping):
+            return list(key_or_list.items())
         elif not isinstance(key_or_list, (list, tuple)):
             raise TypeError("if no direction is specified, key_or_list must be an instance of list")
-        values = []
+        values: list[tuple[str, int]] = []
         for item in key_or_list:
             if isinstance(item, str):
-                item = (item, ASCENDING)
+                item = (item, ASCENDING)  # noqa: PLW2901
             values.append(item)
         return values
 
 
-def _index_document(index_list):
+def _index_document(index_list: _IndexList) -> dict[str, Any]:
     """Helper to generate an index specifying document.
 
     Takes a list of (key, direction) pairs.
     """
-    if isinstance(index_list, abc.Mapping):
+    if not isinstance(index_list, (list, tuple, abc.Mapping)):
         raise TypeError(
-            "passing a dict to sort/create_index/hint is not "
-            "allowed - use a list of tuples instead. did you "
-            "mean %r?" % list(index_list.items())
+            "must use a dictionary or a list of (key, direction) pairs, not: " + repr(index_list)
         )
-    elif not isinstance(index_list, (list, tuple)):
-        raise TypeError("must use a list of (key, direction) pairs, not: " + repr(index_list))
     if not len(index_list):
-        raise ValueError("key_or_list must not be the empty list")
+        raise ValueError("key_or_list must not be empty")
 
-    index: SON[str, Any] = SON()
-    for item in index_list:
-        if isinstance(item, str):
-            item = (item, ASCENDING)
-        key, value = item
-        if not isinstance(key, str):
-            raise TypeError("first item in each key pair must be an instance of str")
-        if not isinstance(value, (str, int, abc.Mapping)):
-            raise TypeError(
-                "second item in each key pair must be 1, -1, "
-                "'2d', or another valid MongoDB index specifier."
-            )
-        index[key] = value
+    index: dict[str, Any] = {}
+
+    if isinstance(index_list, abc.Mapping):
+        for key in index_list:
+            value = index_list[key]
+            _validate_index_key_pair(key, value)
+            index[key] = value
+    else:
+        for item in index_list:
+            if isinstance(item, str):
+                item = (item, ASCENDING)  # noqa: PLW2901
+            key, value = item
+            _validate_index_key_pair(key, value)
+            index[key] = value
     return index
 
 
+def _validate_index_key_pair(key: Any, value: Any) -> None:
+    if not isinstance(key, str):
+        raise TypeError("first item in each key pair must be an instance of str")
+    if not isinstance(value, (str, int, abc.Mapping)):
+        raise TypeError(
+            "second item in each key pair must be 1, -1, "
+            "'2d', or another valid MongoDB index specifier."
+        )
+
+
 def _check_command_response(
-    response, max_wire_version, allowable_errors=None, parse_write_concern_error=False
-):
+    response: _DocumentOut,
+    max_wire_version: Optional[int],
+    allowable_errors: Optional[Container[Union[int, str]]] = None,
+    parse_write_concern_error: bool = False,
+) -> None:
     """Check the response to a command for errors."""
     if "ok" not in response:
         # Server didn't recognize our message as a command.
         raise OperationFailure(
-            response.get("$err"), response.get("code"), response, max_wire_version
+            response.get("$err"),  # type: ignore[arg-type]
+            response.get("code"),
+            response,
+            max_wire_version,
         )
 
     if parse_write_concern_error and "writeConcernError" in response:
@@ -194,7 +233,7 @@ def _check_command_response(
     raise OperationFailure(errmsg, code, response, max_wire_version)
 
 
-def _raise_last_write_error(write_errors: List[Any]) -> NoReturn:
+def _raise_last_write_error(write_errors: list[Any]) -> NoReturn:
     # If the last batch had multiple errors only report
     # the last error to emulate continue_on_error.
     error = write_errors[-1]
@@ -210,7 +249,7 @@ def _raise_write_concern_error(error: Any) -> NoReturn:
     raise WriteConcernError(error.get("errmsg"), error.get("code"), error)
 
 
-def _get_wce_doc(result):
+def _get_wce_doc(result: Mapping[str, Any]) -> Optional[Mapping[str, Any]]:
     """Return the writeConcernError or None."""
     wce = result.get("writeConcernError")
     if wce:
@@ -218,11 +257,13 @@ def _get_wce_doc(result):
         # convenient to attach it to the writeConcernError doc itself.
         error_labels = result.get("errorLabels")
         if error_labels:
+            # Copy to avoid changing the original document.
+            wce = wce.copy()
             wce["errorLabels"] = error_labels
     return wce
 
 
-def _check_write_command_response(result):
+def _check_write_command_response(result: Mapping[str, Any]) -> None:
     """Backward compatibility helper for write command error handling."""
     # Prefer write errors over write concern errors
     write_errors = result.get("writeErrors")
@@ -234,7 +275,9 @@ def _check_write_command_response(result):
         _raise_write_concern_error(wce)
 
 
-def _fields_list_to_dict(fields, option_name):
+def _fields_list_to_dict(
+    fields: Union[Mapping[str, Any], Iterable[str]], option_name: str
+) -> Mapping[str, Any]:
     """Takes a sequence of field names and returns a matching dictionary.
 
     ["a", "b"] becomes {"a": 1, "b": 1}
@@ -254,7 +297,7 @@ def _fields_list_to_dict(fields, option_name):
     raise TypeError(f"{option_name} must be a mapping or list of key names")
 
 
-def _handle_exception():
+def _handle_exception() -> None:
     """Print exceptions raised by subscribers to stderr."""
     # Heavily influenced by logging.Handler.handleError.
 
@@ -270,10 +313,15 @@ def _handle_exception():
             del einfo
 
 
-def _handle_reauth(func):
-    def inner(*args, **kwargs):
+# See https://mypy.readthedocs.io/en/stable/generics.html?#decorator-factories
+F = TypeVar("F", bound=Callable[..., Any])
+
+
+def _handle_reauth(func: F) -> F:
+    def inner(*args: Any, **kwargs: Any) -> Any:
         no_reauth = kwargs.pop("no_reauth", False)
-        from pymongo.pool import SocketInfo
+        from pymongo.message import _BulkWriteContext
+        from pymongo.pool import Connection
 
         try:
             return func(*args, **kwargs)
@@ -281,22 +329,22 @@ def _handle_reauth(func):
             if no_reauth:
                 raise
             if exc.code == _REAUTHENTICATION_REQUIRED_CODE:
-                # Look for an argument that either is a SocketInfo
-                # or has a socket_info attribute, so we can trigger
+                # Look for an argument that either is a Connection
+                # or has a connection attribute, so we can trigger
                 # a reauth.
-                sock_info = None
+                conn = None
                 for arg in args:
-                    if isinstance(arg, SocketInfo):
-                        sock_info = arg
+                    if isinstance(arg, Connection):
+                        conn = arg
                         break
-                    if hasattr(arg, "sock_info"):
-                        sock_info = arg.sock_info
+                    if isinstance(arg, _BulkWriteContext):
+                        conn = arg.conn
                         break
-                if sock_info:
-                    sock_info.authenticate(reauthenticate=True)
+                if conn:
+                    conn.authenticate(reauthenticate=True)
                 else:
                     raise
                 return func(*args, **kwargs)
             raise
 
-    return inner
+    return cast(F, inner)
