@@ -14,18 +14,22 @@
 
 """Internal class to monitor a topology of one or more servers."""
 
+from __future__ import annotations
+
+import logging
 import os
 import queue
 import random
+import sys
 import time
 import warnings
 import weakref
-from typing import Any
+from pathlib import Path
+from typing import TYPE_CHECKING, Any, Callable, Mapping, Optional, cast
 
 from pymongo import _csot, common, helpers, periodic_executor
-from pymongo.client_session import _ServerSessionPool
+from pymongo.client_session import _ServerSession, _ServerSessionPool
 from pymongo.errors import (
-    ConfigurationError,
     ConnectionFailure,
     InvalidOperation,
     NetworkTimeout,
@@ -37,15 +41,20 @@ from pymongo.errors import (
 )
 from pymongo.hello import Hello
 from pymongo.lock import _create_lock
+from pymongo.logger import (
+    _SERVER_SELECTION_LOGGER,
+    _debug_log,
+    _info_log,
+    _ServerSelectionStatusMessage,
+)
 from pymongo.monitor import SrvMonitor
-from pymongo.pool import PoolOptions
+from pymongo.pool import Pool, PoolOptions
 from pymongo.server import Server
 from pymongo.server_description import ServerDescription
 from pymongo.server_selectors import (
     Selection,
     any_server_selector,
     arbiter_server_selector,
-    readable_server_selector,
     secondary_server_selector,
     writable_server_selector,
 )
@@ -57,8 +66,16 @@ from pymongo.topology_description import (
     updated_topology_description,
 )
 
+if TYPE_CHECKING:
+    from bson import ObjectId
+    from pymongo.settings import TopologySettings
+    from pymongo.typings import ClusterTime, _Address
 
-def process_events_queue(queue_ref):
+
+_pymongo_dir = str(Path(__file__).parent)
+
+
+def process_events_queue(queue_ref: weakref.ReferenceType[queue.Queue]) -> bool:
     q = queue_ref()
     if not q:
         return False  # Cancel PeriodicExecutor.
@@ -75,15 +92,14 @@ def process_events_queue(queue_ref):
     return True  # Continue PeriodicExecutor.
 
 
-class Topology(object):
+class Topology:
     """Monitor a topology of one or more servers."""
 
-    def __init__(self, topology_settings):
+    def __init__(self, topology_settings: TopologySettings):
         self._topology_id = topology_settings._topology_id
         self._listeners = topology_settings._pool_options._event_listeners
-        pub = self._listeners is not None
-        self._publish_server = pub and self._listeners.enabled_for_server
-        self._publish_tp = pub and self._listeners.enabled_for_topology
+        self._publish_server = self._listeners is not None and self._listeners.enabled_for_server
+        self._publish_tp = self._listeners is not None and self._listeners.enabled_for_topology
 
         # Create events queue if there are publishers.
         self._events = None
@@ -129,14 +145,16 @@ class Topology(object):
         self._closed = False
         self._lock = _create_lock()
         self._condition = self._settings.condition_class(self._lock)
-        self._servers = {}
-        self._pid = None
-        self._max_cluster_time = None
+        self._servers: dict[_Address, Server] = {}
+        self._pid: Optional[int] = None
+        self._max_cluster_time: Optional[ClusterTime] = None
         self._session_pool = _ServerSessionPool()
 
         if self._publish_server or self._publish_tp:
+            assert self._events is not None
+            weak: weakref.ReferenceType[queue.Queue]
 
-            def target():
+            def target() -> bool:
                 return process_events_queue(weak)
 
             executor = periodic_executor.PeriodicExecutor(
@@ -157,7 +175,7 @@ class Topology(object):
         if self._settings.fqdn is not None and not self._settings.load_balanced:
             self._srv_monitor = SrvMonitor(self, self._settings)
 
-    def open(self):
+    def open(self) -> None:
         """Start monitoring, or restart after a fork.
 
         No effect if called multiple times.
@@ -174,11 +192,17 @@ class Topology(object):
             self._pid = pid
         elif pid != self._pid:
             self._pid = pid
-            warnings.warn(
+            if sys.version_info[:2] >= (3, 12):
+                kwargs = {"skip_file_prefixes": (_pymongo_dir,)}
+            else:
+                kwargs = {"stacklevel": 6}
+            # Ignore B028 warning for missing stacklevel.
+            warnings.warn(  # type: ignore[call-overload] # noqa: B028
                 "MongoClient opened before fork. May not be entirely fork-safe, "
                 "proceed with caution. See PyMongo's documentation for details: "
                 "https://pymongo.readthedocs.io/en/stable/faq.html#"
-                "is-pymongo-fork-safe"
+                "is-pymongo-fork-safe",
+                **kwargs,
             )
             with self._lock:
                 # Close servers and clear the pools.
@@ -191,23 +215,30 @@ class Topology(object):
         with self._lock:
             self._ensure_opened()
 
-    def get_server_selection_timeout(self):
+    def get_server_selection_timeout(self) -> float:
         # CSOT: use remaining timeout when set.
         timeout = _csot.remaining()
         if timeout is None:
             return self._settings.server_selection_timeout
         return timeout
 
-    def select_servers(self, selector, server_selection_timeout=None, address=None):
+    def select_servers(
+        self,
+        selector: Callable[[Selection], Selection],
+        operation: str,
+        server_selection_timeout: Optional[float] = None,
+        address: Optional[_Address] = None,
+        operation_id: Optional[int] = None,
+    ) -> list[Server]:
         """Return a list of Servers matching selector, or time out.
 
-        :Parameters:
-          - `selector`: function that takes a list of Servers and returns
+        :param selector: function that takes a list of Servers and returns
             a subset of them.
-          - `server_selection_timeout` (optional): maximum seconds to wait.
+        :param operation: The name of the operation that the server is being selected for.
+        :param server_selection_timeout: maximum seconds to wait.
             If not provided, the default value common.SERVER_SELECTION_TIMEOUT
             is used.
-          - `address`: optional server address to select.
+        :param address: optional server address to select.
 
         Calls self.open() if needed.
 
@@ -220,14 +251,38 @@ class Topology(object):
             server_timeout = server_selection_timeout
 
         with self._lock:
-            server_descriptions = self._select_servers_loop(selector, server_timeout, address)
+            server_descriptions = self._select_servers_loop(
+                selector, server_timeout, operation, operation_id, address
+            )
 
-            return [self.get_server_by_address(sd.address) for sd in server_descriptions]
+            return [
+                cast(Server, self.get_server_by_address(sd.address)) for sd in server_descriptions
+            ]
 
-    def _select_servers_loop(self, selector, timeout, address):
+    def _select_servers_loop(
+        self,
+        selector: Callable[[Selection], Selection],
+        timeout: float,
+        operation: str,
+        operation_id: Optional[int],
+        address: Optional[_Address],
+    ) -> list[ServerDescription]:
         """select_servers() guts. Hold the lock when calling this."""
         now = time.monotonic()
         end_time = now + timeout
+        logged_waiting = False
+
+        if _SERVER_SELECTION_LOGGER.isEnabledFor(logging.DEBUG):
+            _debug_log(
+                _SERVER_SELECTION_LOGGER,
+                message=_ServerSelectionStatusMessage.STARTED,
+                selector=selector,
+                operation=operation,
+                operationId=operation_id,
+                topologyDescription=self.description,
+                clientId=self.description._topology_settings._topology_id,
+            )
+
         server_descriptions = self._description.apply_selector(
             selector, address, custom_selector=self._settings.server_selector
         )
@@ -235,11 +290,34 @@ class Topology(object):
         while not server_descriptions:
             # No suitable servers.
             if timeout == 0 or now > end_time:
+                if _SERVER_SELECTION_LOGGER.isEnabledFor(logging.DEBUG):
+                    _debug_log(
+                        _SERVER_SELECTION_LOGGER,
+                        message=_ServerSelectionStatusMessage.FAILED,
+                        selector=selector,
+                        operation=operation,
+                        operationId=operation_id,
+                        topologyDescription=self.description,
+                        clientId=self.description._topology_settings._topology_id,
+                        failure=self._error_message(selector),
+                    )
                 raise ServerSelectionTimeoutError(
-                    "%s, Timeout: %ss, Topology Description: %r"
-                    % (self._error_message(selector), timeout, self.description)
+                    f"{self._error_message(selector)}, Timeout: {timeout}s, Topology Description: {self.description!r}"
                 )
 
+            if not logged_waiting:
+                _info_log(
+                    _SERVER_SELECTION_LOGGER,
+                    message=_ServerSelectionStatusMessage.WAITING,
+                    selector=selector,
+                    operation=operation,
+                    operationId=operation_id,
+                    topologyDescription=self.description,
+                    clientId=self.description._topology_settings._topology_id,
+                    remainingTimeMS=int(end_time - time.monotonic()),
+                )
+                logged_waiting = True
+
             self._ensure_opened()
             self._request_check_all()
 
@@ -257,8 +335,19 @@ class Topology(object):
         self._description.check_compatible()
         return server_descriptions
 
-    def _select_server(self, selector, server_selection_timeout=None, address=None):
-        servers = self.select_servers(selector, server_selection_timeout, address)
+    def _select_server(
+        self,
+        selector: Callable[[Selection], Selection],
+        operation: str,
+        server_selection_timeout: Optional[float] = None,
+        address: Optional[_Address] = None,
+        deprioritized_servers: Optional[list[Server]] = None,
+        operation_id: Optional[int] = None,
+    ) -> Server:
+        servers = self.select_servers(
+            selector, operation, server_selection_timeout, address, operation_id
+        )
+        servers = _filter_servers(servers, deprioritized_servers)
         if len(servers) == 1:
             return servers[0]
         server1, server2 = random.sample(servers, 2)
@@ -267,34 +356,79 @@ class Topology(object):
         else:
             return server2
 
-    def select_server(self, selector, server_selection_timeout=None, address=None):
+    def select_server(
+        self,
+        selector: Callable[[Selection], Selection],
+        operation: str,
+        server_selection_timeout: Optional[float] = None,
+        address: Optional[_Address] = None,
+        deprioritized_servers: Optional[list[Server]] = None,
+        operation_id: Optional[int] = None,
+    ) -> Server:
         """Like select_servers, but choose a random server if several match."""
-        server = self._select_server(selector, server_selection_timeout, address)
+        server = self._select_server(
+            selector,
+            operation,
+            server_selection_timeout,
+            address,
+            deprioritized_servers,
+            operation_id=operation_id,
+        )
         if _csot.get_timeout():
-            _csot.set_rtt(server.description.round_trip_time)
+            _csot.set_rtt(server.description.min_round_trip_time)
+        if _SERVER_SELECTION_LOGGER.isEnabledFor(logging.DEBUG):
+            _debug_log(
+                _SERVER_SELECTION_LOGGER,
+                message=_ServerSelectionStatusMessage.SUCCEEDED,
+                selector=selector,
+                operation=operation,
+                operationId=operation_id,
+                topologyDescription=self.description,
+                clientId=self.description._topology_settings._topology_id,
+                serverHost=server.description.address[0],
+                serverPort=server.description.address[1],
+            )
         return server
 
-    def select_server_by_address(self, address, server_selection_timeout=None):
+    def select_server_by_address(
+        self,
+        address: _Address,
+        operation: str,
+        server_selection_timeout: Optional[int] = None,
+        operation_id: Optional[int] = None,
+    ) -> Server:
         """Return a Server for "address", reconnecting if necessary.
 
         If the server's type is not known, request an immediate check of all
         servers. Time out after "server_selection_timeout" if the server
         cannot be reached.
 
-        :Parameters:
-          - `address`: A (host, port) pair.
-          - `server_selection_timeout` (optional): maximum seconds to wait.
+        :param address: A (host, port) pair.
+        :param operation: The name of the operation that the server is being selected for.
+        :param server_selection_timeout: maximum seconds to wait.
             If not provided, the default value
             common.SERVER_SELECTION_TIMEOUT is used.
+        :param operation_id: The unique id of the current operation being performed. Defaults to None if not provided.
 
         Calls self.open() if needed.
 
         Raises exc:`ServerSelectionTimeoutError` after
         `server_selection_timeout` if no matching servers are found.
         """
-        return self.select_server(any_server_selector, server_selection_timeout, address)
+        return self.select_server(
+            any_server_selector,
+            operation,
+            server_selection_timeout,
+            address,
+            operation_id=operation_id,
+        )
 
-    def _process_change(self, server_description, reset_pool=False):
+    def _process_change(
+        self,
+        server_description: ServerDescription,
+        reset_pool: bool = False,
+        interrupt_connections: bool = False,
+    ) -> None:
         """Process a new ServerDescription on an opened topology.
 
         Hold the lock when calling this.
@@ -350,12 +484,17 @@ class Topology(object):
         if reset_pool:
             server = self._servers.get(server_description.address)
             if server:
-                server.pool.reset()
+                server.pool.reset(interrupt_connections=interrupt_connections)
 
         # Wake waiters in select_servers().
         self._condition.notify_all()
 
-    def on_change(self, server_description, reset_pool=False):
+    def on_change(
+        self,
+        server_description: ServerDescription,
+        reset_pool: bool = False,
+        interrupt_connections: bool = False,
+    ) -> None:
         """Process a new ServerDescription after an hello call completes."""
         # We do no I/O holding the lock.
         with self._lock:
@@ -368,9 +507,9 @@ class Topology(object):
             # change removed it. E.g., we got a host list from the primary
             # that didn't include this server.
             if self._opened and self._description.has_server(server_description.address):
-                self._process_change(server_description, reset_pool)
+                self._process_change(server_description, reset_pool, interrupt_connections)
 
-    def _process_srv_update(self, seedlist):
+    def _process_srv_update(self, seedlist: list[tuple[str, Any]]) -> None:
         """Process a new seedlist on an opened topology.
         Hold the lock when calling this.
         """
@@ -390,14 +529,14 @@ class Topology(object):
                 )
             )
 
-    def on_srv_update(self, seedlist):
+    def on_srv_update(self, seedlist: list[tuple[str, Any]]) -> None:
         """Process a new list of nodes obtained from scanning SRV records."""
         # We do no I/O holding the lock.
         with self._lock:
             if self._opened:
                 self._process_srv_update(seedlist)
 
-    def get_server_by_address(self, address):
+    def get_server_by_address(self, address: _Address) -> Optional[Server]:
         """Get a Server or None.
 
         Returns the current version of the server immediately, even if it's
@@ -407,10 +546,10 @@ class Topology(object):
         """
         return self._servers.get(address)
 
-    def has_server(self, address):
+    def has_server(self, address: _Address) -> bool:
         return address in self._servers
 
-    def get_primary(self):
+    def get_primary(self) -> Optional[_Address]:
         """Return primary's address or None."""
         # Implemented here in Topology instead of MongoClient, so it can lock.
         with self._lock:
@@ -420,7 +559,7 @@ class Topology(object):
 
             return writable_server_selector(self._new_selection())[0].address
 
-    def _get_replica_set_members(self, selector):
+    def _get_replica_set_members(self, selector: Callable[[Selection], Selection]) -> set[_Address]:
         """Return set of replica set member addresses."""
         # Implemented here in Topology instead of MongoClient, so it can lock.
         with self._lock:
@@ -431,21 +570,21 @@ class Topology(object):
             ):
                 return set()
 
-            return set([sd.address for sd in selector(self._new_selection())])
+            return {sd.address for sd in iter(selector(self._new_selection()))}
 
-    def get_secondaries(self):
+    def get_secondaries(self) -> set[_Address]:
         """Return set of secondary addresses."""
         return self._get_replica_set_members(secondary_server_selector)
 
-    def get_arbiters(self):
+    def get_arbiters(self) -> set[_Address]:
         """Return set of arbiter addresses."""
         return self._get_replica_set_members(arbiter_server_selector)
 
-    def max_cluster_time(self):
+    def max_cluster_time(self) -> Optional[ClusterTime]:
         """Return a document, the highest seen $clusterTime."""
         return self._max_cluster_time
 
-    def _receive_cluster_time_no_lock(self, cluster_time):
+    def _receive_cluster_time_no_lock(self, cluster_time: Optional[Mapping[str, Any]]) -> None:
         # Driver Sessions Spec: "Whenever a driver receives a cluster time from
         # a server it MUST compare it to the current highest seen cluster time
         # for the deployment. If the new cluster time is higher than the
@@ -460,17 +599,17 @@ class Topology(object):
             ):
                 self._max_cluster_time = cluster_time
 
-    def receive_cluster_time(self, cluster_time):
+    def receive_cluster_time(self, cluster_time: Optional[Mapping[str, Any]]) -> None:
         with self._lock:
             self._receive_cluster_time_no_lock(cluster_time)
 
-    def request_check_all(self, wait_time=5):
+    def request_check_all(self, wait_time: int = 5) -> None:
         """Wake all monitors, wait for at least one to check its server."""
         with self._lock:
             self._request_check_all()
             self._condition.wait(wait_time)
 
-    def data_bearing_servers(self):
+    def data_bearing_servers(self) -> list[ServerDescription]:
         """Return a list of all data-bearing servers.
 
         This includes any server that might be selected for an operation.
@@ -479,7 +618,7 @@ class Topology(object):
             return self._description.known_servers
         return self._description.readable_servers
 
-    def update_pool(self):
+    def update_pool(self) -> None:
         # Remove any stale sockets and add new sockets if pool is too small.
         servers = []
         with self._lock:
@@ -496,10 +635,11 @@ class Topology(object):
                 self.handle_error(server.description.address, ctx)
                 raise
 
-    def close(self):
+    def close(self) -> None:
         """Clear pools and terminate monitors. Topology does not reopen on
         demand. Any further operations will raise
-        :exc:`~.errors.InvalidOperation`."""
+        :exc:`~.errors.InvalidOperation`.
+        """
         with self._lock:
             for server in self._servers.values():
                 server.close()
@@ -525,48 +665,20 @@ class Topology(object):
             self.__events_executor.close()
 
     @property
-    def description(self):
+    def description(self) -> TopologyDescription:
         return self._description
 
-    def pop_all_sessions(self):
+    def pop_all_sessions(self) -> list[_ServerSession]:
         """Pop all session ids from the pool."""
         with self._lock:
             return self._session_pool.pop_all()
 
-    def _check_implicit_session_support(self):
-        with self._lock:
-            self._check_session_support()
-
-    def _check_session_support(self):
-        """Internal check for session support on clusters."""
-        if self._settings.load_balanced:
-            # Sessions never time out in load balanced mode.
-            return float("inf")
-        session_timeout = self._description.logical_session_timeout_minutes
-        if session_timeout is None:
-            # Maybe we need an initial scan? Can raise ServerSelectionError.
-            if self._description.topology_type == TOPOLOGY_TYPE.Single:
-                if not self._description.has_known_servers:
-                    self._select_servers_loop(
-                        any_server_selector, self.get_server_selection_timeout(), None
-                    )
-            elif not self._description.readable_servers:
-                self._select_servers_loop(
-                    readable_server_selector, self.get_server_selection_timeout(), None
-                )
-
-            session_timeout = self._description.logical_session_timeout_minutes
-            if session_timeout is None:
-                raise ConfigurationError("Sessions are not supported by this MongoDB deployment")
-        return session_timeout
-
-    def get_server_session(self):
+    def get_server_session(self, session_timeout_minutes: Optional[int]) -> _ServerSession:
         """Start or resume a server session, or raise ConfigurationError."""
         with self._lock:
-            session_timeout = self._check_session_support()
-            return self._session_pool.get_server_session(session_timeout)
+            return self._session_pool.get_server_session(session_timeout_minutes)
 
-    def return_server_session(self, server_session, lock):
+    def return_server_session(self, server_session: _ServerSession, lock: bool) -> None:
         if lock:
             with self._lock:
                 self._session_pool.return_server_session(
@@ -576,14 +688,14 @@ class Topology(object):
             # Called from a __del__ method, can't use a lock.
             self._session_pool.return_server_session_no_lock(server_session)
 
-    def _new_selection(self):
+    def _new_selection(self) -> Selection:
         """A Selection object, initially including all known servers.
 
         Hold the lock when calling this.
         """
         return Selection.from_topology_description(self._description)
 
-    def _ensure_opened(self):
+    def _ensure_opened(self) -> None:
         """Start monitors, or restart after a fork.
 
         Hold the lock when calling this.
@@ -616,7 +728,7 @@ class Topology(object):
         for server in self._servers.values():
             server.open()
 
-    def _is_stale_error(self, address, err_ctx):
+    def _is_stale_error(self, address: _Address, err_ctx: _ErrorContext) -> bool:
         server = self._servers.get(address)
         if server is None:
             # Another thread removed this server from the topology.
@@ -636,13 +748,12 @@ class Topology(object):
 
         return _is_stale_error_topology_version(cur_tv, error_tv)
 
-    def _handle_error(self, address, err_ctx):
+    def _handle_error(self, address: _Address, err_ctx: _ErrorContext) -> None:
         if self._is_stale_error(address, err_ctx):
             return
 
         server = self._servers[address]
         error = err_ctx.error
-        exc_type = type(error)
         service_id = err_ctx.service_id
 
         # Ignore a handshake error if the server is behind a load balancer but
@@ -652,16 +763,16 @@ class Topology(object):
         if self._settings.load_balanced and not service_id and not err_ctx.completed_handshake:
             return
 
-        if issubclass(exc_type, NetworkTimeout) and err_ctx.completed_handshake:
+        if isinstance(error, NetworkTimeout) and err_ctx.completed_handshake:
             # The socket has been closed. Don't reset the server.
             # Server Discovery And Monitoring Spec: "When an application
             # operation fails because of any network error besides a socket
             # timeout...."
             return
-        elif issubclass(exc_type, WriteError):
+        elif isinstance(error, WriteError):
             # Ignore writeErrors.
             return
-        elif issubclass(exc_type, (NotPrimaryError, OperationFailure)):
+        elif isinstance(error, (NotPrimaryError, OperationFailure)):
             # As per the SDAM spec if:
             #   - the server sees a "not primary" error, and
             #   - the server is not shutting down, and
@@ -675,7 +786,7 @@ class Topology(object):
             else:
                 # Default error code if one does not exist.
                 default = 10107 if isinstance(error, NotPrimaryError) else None
-                err_code = error.details.get("code", default)
+                err_code = error.details.get("code", default)  # type: ignore[union-attr]
             if err_code in helpers._NOT_PRIMARY_CODES:
                 is_shutting_down = err_code in helpers._SHUTDOWN_CODES
                 # Mark server Unknown, clear the pool, and request check.
@@ -691,7 +802,7 @@ class Topology(object):
                     self._process_change(ServerDescription(address, error=error))
                 # Clear the pool.
                 server.reset(service_id)
-        elif issubclass(exc_type, ConnectionFailure):
+        elif isinstance(error, ConnectionFailure):
             # "Client MUST replace the server's description with type Unknown
             # ... MUST NOT request an immediate check of the server."
             if not self._settings.load_balanced:
@@ -703,7 +814,7 @@ class Topology(object):
             # that server and close the current monitoring connection."
             server._monitor.cancel_check()
 
-    def handle_error(self, address, err_ctx):
+    def handle_error(self, address: _Address, err_ctx: _ErrorContext) -> None:
         """Handle an application error.
 
         May reset the server to Unknown, clear the pool, and request an
@@ -712,12 +823,12 @@ class Topology(object):
         with self._lock:
             self._handle_error(address, err_ctx)
 
-    def _request_check_all(self):
+    def _request_check_all(self) -> None:
         """Wake all monitors. Hold the lock when calling this."""
         for server in self._servers.values():
             server.request_check()
 
-    def _update_servers(self):
+    def _update_servers(self) -> None:
         """Sync our Servers from TopologyDescription.server_descriptions.
 
         Hold the lock while calling this.
@@ -732,7 +843,7 @@ class Topology(object):
                 )
 
                 weak = None
-                if self._publish_server:
+                if self._publish_server and self._events is not None:
                     weak = weakref.ref(self._events)
                 server = Server(
                     server_description=sd,
@@ -759,10 +870,12 @@ class Topology(object):
                 server.close()
                 self._servers.pop(address)
 
-    def _create_pool_for_server(self, address):
-        return self._settings.pool_class(address, self._settings.pool_options)
+    def _create_pool_for_server(self, address: _Address) -> Pool:
+        return self._settings.pool_class(
+            address, self._settings.pool_options, client_id=self._topology_id
+        )
 
-    def _create_pool_for_monitor(self, address):
+    def _create_pool_for_monitor(self, address: _Address) -> Pool:
         options = self._settings.pool_options
 
         # According to the Server Discovery And Monitoring Spec, monitors use
@@ -780,9 +893,11 @@ class Topology(object):
             server_api=options.server_api,
         )
 
-        return self._settings.pool_class(address, monitor_pool_options, handshake=False)
+        return self._settings.pool_class(
+            address, monitor_pool_options, handshake=False, client_id=self._topology_id
+        )
 
-    def _error_message(self, selector):
+    def _error_message(self, selector: Callable[[Selection], Selection]) -> str:
         """Format an error message if server selection fails.
 
         Hold the lock when calling this.
@@ -807,14 +922,14 @@ class Topology(object):
                 else:
                     return "No %s available for writes" % server_plural
             else:
-                return 'No %s match selector "%s"' % (server_plural, selector)
+                return f'No {server_plural} match selector "{selector}"'
         else:
             addresses = list(self._description.server_descriptions())
             servers = list(self._description.server_descriptions().values())
             if not servers:
                 if is_replica_set:
                     # We removed all servers because of the wrong setName?
-                    return 'No %s available for replica set name "%s"' % (
+                    return 'No {} available for replica set name "{}"'.format(
                         server_plural,
                         self._settings.replica_set_name,
                     )
@@ -840,30 +955,37 @@ class Topology(object):
             else:
                 return ",".join(str(server.error) for server in servers if server.error)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         msg = ""
         if not self._opened:
             msg = "CLOSED "
-        return "<%s %s%r>" % (self.__class__.__name__, msg, self._description)
+        return f"<{self.__class__.__name__} {msg}{self._description!r}>"
 
-    def eq_props(self):
+    def eq_props(self) -> tuple[tuple[_Address, ...], Optional[str], Optional[str], str]:
         """The properties to use for MongoClient/Topology equality checks."""
         ts = self._settings
         return (tuple(sorted(ts.seeds)), ts.replica_set_name, ts.fqdn, ts.srv_service_name)
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         if isinstance(other, self.__class__):
             return self.eq_props() == other.eq_props()
         return NotImplemented
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash(self.eq_props())
 
 
-class _ErrorContext(object):
+class _ErrorContext:
     """An error with context for SDAM error handling."""
 
-    def __init__(self, error, max_wire_version, sock_generation, completed_handshake, service_id):
+    def __init__(
+        self,
+        error: BaseException,
+        max_wire_version: int,
+        sock_generation: int,
+        completed_handshake: bool,
+        service_id: Optional[ObjectId],
+    ):
         self.error = error
         self.max_wire_version = max_wire_version
         self.sock_generation = sock_generation
@@ -871,7 +993,9 @@ class _ErrorContext(object):
         self.service_id = service_id
 
 
-def _is_stale_error_topology_version(current_tv, error_tv):
+def _is_stale_error_topology_version(
+    current_tv: Optional[Mapping[str, Any]], error_tv: Optional[Mapping[str, Any]]
+) -> bool:
     """Return True if the error's topologyVersion is <= current."""
     if current_tv is None or error_tv is None:
         return False
@@ -880,7 +1004,7 @@ def _is_stale_error_topology_version(current_tv, error_tv):
     return current_tv["counter"] >= error_tv["counter"]
 
 
-def _is_stale_server_description(current_sd, new_sd):
+def _is_stale_server_description(current_sd: ServerDescription, new_sd: ServerDescription) -> bool:
     """Return True if the new topologyVersion is < current."""
     current_tv, new_tv = current_sd.topology_version, new_sd.topology_version
     if current_tv is None or new_tv is None:
@@ -888,3 +1012,16 @@ def _is_stale_server_description(current_sd, new_sd):
     if current_tv["processId"] != new_tv["processId"]:
         return False
     return current_tv["counter"] > new_tv["counter"]
+
+
+def _filter_servers(
+    candidates: list[Server], deprioritized_servers: Optional[list[Server]] = None
+) -> list[Server]:
+    """Filter out deprioritized servers from a list of server candidates."""
+    if not deprioritized_servers:
+        return candidates
+
+    filtered = [server for server in candidates if server not in deprioritized_servers]
+
+    # If not possible to pick a prioritized server, return the original list
+    return filtered or candidates
