@@ -14,10 +14,21 @@
 
 
 """Tools to parse and validate a MongoDB URI."""
+from __future__ import annotations
+
 import re
 import sys
 import warnings
-from typing import Any, Dict, List, Mapping, MutableMapping, Optional, Tuple, Union
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Mapping,
+    MutableMapping,
+    Optional,
+    Sized,
+    Union,
+    cast,
+)
 from urllib.parse import unquote_plus
 
 from pymongo.client_options import _parse_ssl_options
@@ -32,6 +43,9 @@ from pymongo.errors import ConfigurationError, InvalidURI
 from pymongo.srv_resolver import _HAVE_DNSPYTHON, _SrvResolver
 from pymongo.typings import _Address
 
+if TYPE_CHECKING:
+    from pymongo.pyopenssl_context import SSLContext
+
 SCHEME = "mongodb://"
 SCHEME_LEN = len(SCHEME)
 SRV_SCHEME = "mongodb+srv://"
@@ -39,11 +53,10 @@ SRV_SCHEME_LEN = len(SRV_SCHEME)
 DEFAULT_PORT = 27017
 
 
-def _unquoted_percent(s):
+def _unquoted_percent(s: str) -> bool:
     """Check for unescaped percent signs.
 
-    :Parameters:
-        - `s`: A string. `s` can have things like '%25', '%2525',
+    :param s: A string. `s` can have things like '%25', '%2525',
            and '%E2%85%A8' but cannot have unquoted percent like '%foo'.
     """
     for i in range(len(s)):
@@ -56,7 +69,7 @@ def _unquoted_percent(s):
     return False
 
 
-def parse_userinfo(userinfo: str) -> Tuple[str, str]:
+def parse_userinfo(userinfo: str) -> tuple[str, str]:
     """Validates the format of user information in a MongoDB URI.
     Reserved characters that are gen-delimiters (":", "/", "?", "#", "[",
     "]", "@") as per RFC 3986 must be escaped.
@@ -64,8 +77,7 @@ def parse_userinfo(userinfo: str) -> Tuple[str, str]:
     Returns a 2-tuple containing the unescaped username followed
     by the unescaped password.
 
-    :Parameters:
-        - `userinfo`: A string of the form <username>:<password>
+    :param userinfo: A string of the form <username>:<password>
     """
     if "@" in userinfo or userinfo.count(":") > 1 or _unquoted_percent(userinfo):
         raise InvalidURI(
@@ -83,16 +95,15 @@ def parse_userinfo(userinfo: str) -> Tuple[str, str]:
 
 def parse_ipv6_literal_host(
     entity: str, default_port: Optional[int]
-) -> Tuple[str, Optional[Union[str, int]]]:
+) -> tuple[str, Optional[Union[str, int]]]:
     """Validates an IPv6 literal host:port string.
 
     Returns a 2-tuple of IPv6 literal followed by port where
     port is default_port if it wasn't specified in entity.
 
-    :Parameters:
-        - `entity`: A string that represents an IPv6 literal enclosed
+    :param entity: A string that represents an IPv6 literal enclosed
                     in braces (e.g. '[::1]' or '[::1]:27017').
-        - `default_port`: The port number to use when one wasn't
+    :param default_port: The port number to use when one wasn't
                           specified in entity.
     """
     if entity.find("]") == -1:
@@ -111,10 +122,9 @@ def parse_host(entity: str, default_port: Optional[int] = DEFAULT_PORT) -> _Addr
     Returns a 2-tuple of host followed by port where port is default_port
     if it wasn't specified in the string.
 
-    :Parameters:
-        - `entity`: A host or host:port string where host could be a
+    :param entity: A host or host:port string where host could be a
                     hostname or IP address.
-        - `default_port`: The port number to use when one wasn't
+    :param default_port: The port number to use when one wasn't
                           specified in entity.
     """
     host = entity
@@ -152,7 +162,7 @@ _IMPLICIT_TLSINSECURE_OPTS = {
 }
 
 
-def _parse_options(opts, delim):
+def _parse_options(opts: str, delim: Optional[str]) -> _CaseInsensitiveDictionary:
     """Helper method for split_options which creates the options dict.
     Also handles the creation of a list for the URI tag_sets/
     readpreferencetags portion, and the use of a unicode options string.
@@ -164,7 +174,7 @@ def _parse_options(opts, delim):
             options.setdefault(key, []).append(value)
         else:
             if key in options:
-                warnings.warn(f"Duplicate URI option '{key}'.")
+                warnings.warn(f"Duplicate URI option '{key}'.", stacklevel=2)
             if key.lower() == "authmechanismproperties":
                 val = value
             else:
@@ -174,12 +184,11 @@ def _parse_options(opts, delim):
     return options
 
 
-def _handle_security_options(options):
+def _handle_security_options(options: _CaseInsensitiveDictionary) -> _CaseInsensitiveDictionary:
     """Raise appropriate errors when conflicting TLS options are present in
     the options dictionary.
 
-    :Parameters:
-        - `options`: Instance of _CaseInsensitiveDictionary containing
+    :param options: Instance of _CaseInsensitiveDictionary containing
           MongoDB URI options.
     """
     # Implicitly defined options must not be explicitly specified.
@@ -214,7 +223,7 @@ def _handle_security_options(options):
 
     if "ssl" in options and "tls" in options:
 
-        def truth_value(val):
+        def truth_value(val: Any) -> Any:
             if val in ("true", "false"):
                 return val == "true"
             if isinstance(val, bool):
@@ -228,13 +237,12 @@ def _handle_security_options(options):
     return options
 
 
-def _handle_option_deprecations(options):
+def _handle_option_deprecations(options: _CaseInsensitiveDictionary) -> _CaseInsensitiveDictionary:
     """Issue appropriate warnings when deprecated options are present in the
     options dictionary. Removes deprecated option key, value pairs if the
     options dictionary is found to also have the renamed option.
 
-    :Parameters:
-        - `options`: Instance of _CaseInsensitiveDictionary containing
+    :param options: Instance of _CaseInsensitiveDictionary containing
           MongoDB URI options.
     """
     for optname in list(options):
@@ -268,12 +276,11 @@ def _handle_option_deprecations(options):
     return options
 
 
-def _normalize_options(options):
+def _normalize_options(options: _CaseInsensitiveDictionary) -> _CaseInsensitiveDictionary:
     """Normalizes option names in the options dictionary by converting them to
     their internally-used names.
 
-    :Parameters:
-        - `options`: Instance of _CaseInsensitiveDictionary containing
+    :param options: Instance of _CaseInsensitiveDictionary containing
           MongoDB URI options.
     """
     # Expand the tlsInsecure option.
@@ -298,9 +305,8 @@ def validate_options(opts: Mapping[str, Any], warn: bool = False) -> MutableMapp
     False then errors will be thrown for invalid options, otherwise they will
     be ignored and a warning will be issued.
 
-    :Parameters:
-        - `opts`: A dict of MongoDB URI options.
-        - `warn` (optional): If ``True`` then warnings will be logged and
+    :param opts: A dict of MongoDB URI options.
+    :param warn: If ``True`` then warnings will be logged and
           invalid options will be ignored. Otherwise invalid options will
           cause errors.
     """
@@ -313,13 +319,12 @@ def split_options(
     """Takes the options portion of a MongoDB URI, validates each option
     and returns the options in a dictionary.
 
-    :Parameters:
-        - `opt`: A string representing MongoDB URI options.
-        - `validate`: If ``True`` (the default), validate and normalize all
+    :param opt: A string representing MongoDB URI options.
+    :param validate: If ``True`` (the default), validate and normalize all
           options.
-        - `warn`: If ``False`` (the default), suppress all warnings raised
+    :param warn: If ``False`` (the default), suppress all warnings raised
           during validation of options.
-        - `normalize`: If ``True`` (the default), renames all options to their
+    :param normalize: If ``True`` (the default), renames all options to their
           internally-used names.
     """
     and_idx = opts.find("&")
@@ -336,7 +341,7 @@ def split_options(
         else:
             raise ValueError
     except ValueError:
-        raise InvalidURI("MongoDB URI options are key=value pairs.")
+        raise InvalidURI("MongoDB URI options are key=value pairs.") from None
 
     options = _handle_security_options(options)
 
@@ -346,14 +351,14 @@ def split_options(
         options = _normalize_options(options)
 
     if validate:
-        options = validate_options(options, warn)
+        options = cast(_CaseInsensitiveDictionary, validate_options(options, warn))
         if options.get("authsource") == "":
             raise InvalidURI("the authSource database cannot be an empty string")
 
     return options
 
 
-def split_hosts(hosts: str, default_port: Optional[int] = DEFAULT_PORT) -> List[_Address]:
+def split_hosts(hosts: str, default_port: Optional[int] = DEFAULT_PORT) -> list[_Address]:
     """Takes a string of the form host1[:port],host2[:port]... and
     splits it into (host, port) tuples. If [:port] isn't present the
     default_port is used.
@@ -361,9 +366,8 @@ def split_hosts(hosts: str, default_port: Optional[int] = DEFAULT_PORT) -> List[
     Returns a set of 2-tuples containing the host name (or IP) followed by
     port number.
 
-    :Parameters:
-        - `hosts`: A string of the form host1[:port],host2[:port],...
-        - `default_port`: The port number to use when one wasn't specified
+    :param hosts: A string of the form host1[:port],host2[:port],...
+    :param default_port: The port number to use when one wasn't specified
           for a host.
     """
     nodes = []
@@ -387,7 +391,7 @@ _ALLOWED_TXT_OPTS = frozenset(
 )
 
 
-def _check_options(nodes, options):
+def _check_options(nodes: Sized, options: Mapping[str, Any]) -> None:
     # Ensure directConnection was not True if there are multiple seeds.
     if len(nodes) > 1 and options.get("directconnection"):
         raise ConfigurationError("Cannot specify multiple hosts with directConnection=true")
@@ -410,7 +414,7 @@ def parse_uri(
     connect_timeout: Optional[float] = None,
     srv_service_name: Optional[str] = None,
     srv_max_hosts: Optional[int] = None,
-) -> Dict[str, Any]:
+) -> dict[str, Any]:
     """Parse and validate a MongoDB URI.
 
     Returns a dict of the form::
@@ -428,21 +432,24 @@ def parse_uri(
     If the URI scheme is "mongodb+srv://" DNS SRV and TXT lookups will be done
     to build nodelist and options.
 
-    :Parameters:
-        - `uri`: The MongoDB URI to parse.
-        - `default_port`: The port number to use when one wasn't specified
+    :param uri: The MongoDB URI to parse.
+    :param default_port: The port number to use when one wasn't specified
           for a host in the URI.
-        - `validate` (optional): If ``True`` (the default), validate and
+    :param validate: If ``True`` (the default), validate and
           normalize all options. Default: ``True``.
-        - `warn` (optional): When validating, if ``True`` then will warn
+    :param warn: When validating, if ``True`` then will warn
           the user then ignore any invalid options or values. If ``False``,
           validation will error when options are unsupported or values are
           invalid. Default: ``False``.
-        - `normalize` (optional): If ``True``, convert names of URI options
+    :param normalize: If ``True``, convert names of URI options
           to their internally-used names. Default: ``True``.
-        - `connect_timeout` (optional): The maximum time in milliseconds to
+    :param connect_timeout: The maximum time in milliseconds to
           wait for a response from the DNS server.
-        - 'srv_service_name` (optional): A custom SRV service name
+    :param srv_service_name: A custom SRV service name
+
+    .. versionchanged:: 4.6
+       The delimiting slash (``/``) between hosts and connection options is now optional.
+       For example, "mongodb://example.com?tls=true" is now a valid URI.
 
     .. versionchanged:: 4.0
        To better follow RFC 3986, unquoted percent signs ("%") are no longer
@@ -492,22 +499,23 @@ def parse_uri(
         host_part = path_part
         path_part = ""
 
-    if not path_part and "?" in host_part:
-        raise InvalidURI("A '/' is required between the host list and any options.")
-
     if path_part:
         dbase, _, opts = path_part.partition("?")
-        if dbase:
-            dbase = unquote_plus(dbase)
-            if "." in dbase:
-                dbase, collection = dbase.split(".", 1)
-            if _BAD_DB_CHARS.search(dbase):
-                raise InvalidURI('Bad database name "%s"' % dbase)
-        else:
-            dbase = None
+    else:
+        # There was no slash in scheme_free, check for a sole "?".
+        host_part, _, opts = host_part.partition("?")
 
-        if opts:
-            options.update(split_options(opts, validate, warn, normalize))
+    if dbase:
+        dbase = unquote_plus(dbase)
+        if "." in dbase:
+            dbase, collection = dbase.split(".", 1)
+        if _BAD_DB_CHARS.search(dbase):
+            raise InvalidURI('Bad database name "%s"' % dbase)
+    else:
+        dbase = None
+
+    if opts:
+        options.update(split_options(opts, validate, warn, normalize))
     if srv_service_name is None:
         srv_service_name = options.get("srvServiceName", SRV_SERVICE_NAME)
     if "@" in host_part:
@@ -577,21 +585,21 @@ def parse_uri(
     }
 
 
-def _parse_kms_tls_options(kms_tls_options):
+def _parse_kms_tls_options(kms_tls_options: Optional[Mapping[str, Any]]) -> dict[str, SSLContext]:
     """Parse KMS TLS connection options."""
     if not kms_tls_options:
         return {}
     if not isinstance(kms_tls_options, dict):
         raise TypeError("kms_tls_options must be a dict")
     contexts = {}
-    for provider, opts in kms_tls_options.items():
-        if not isinstance(opts, dict):
+    for provider, options in kms_tls_options.items():
+        if not isinstance(options, dict):
             raise TypeError(f'kms_tls_options["{provider}"] must be a dict')
-        opts.setdefault("tls", True)
-        opts = _CaseInsensitiveDictionary(opts)
+        options.setdefault("tls", True)
+        opts = _CaseInsensitiveDictionary(options)
         opts = _handle_security_options(opts)
         opts = _normalize_options(opts)
-        opts = validate_options(opts)
+        opts = cast(_CaseInsensitiveDictionary, validate_options(opts))
         ssl_context, allow_invalid_hostnames = _parse_ssl_options(opts)
         if ssl_context is None:
             raise ConfigurationError("TLS is required for KMS providers")
@@ -614,7 +622,7 @@ if __name__ == "__main__":
     import pprint
 
     try:
-        pprint.pprint(parse_uri(sys.argv[1]))
+        pprint.pprint(parse_uri(sys.argv[1]))  # noqa: T203
     except InvalidURI as exc:
-        print(exc)
+        print(exc)  # noqa: T201
     sys.exit(0)
