@@ -51,7 +51,7 @@ from pymongo.database import Database
 from pymongo.hello import HelloCompat
 from pymongo.mongo_client import MongoClient
 from pymongo.server_api import ServerApi
-from pymongo.ssl_support import HAVE_SSL, _ssl
+from pymongo.ssl_support import HAVE_SSL, _ssl  # type:ignore[attr-defined]
 from pymongo.uri_parser import parse_uri
 
 if HAVE_SSL:
@@ -67,6 +67,7 @@ if hasattr(gc, "set_debug"):
 # for a replica set.
 host = os.environ.get("DB_IP", "localhost")
 port = int(os.environ.get("DB_PORT", 27017))
+IS_SRV = "mongodb+srv" in host
 
 db_user = os.environ.get("DB_USER", "user")
 db_pwd = os.environ.get("DB_PASSWORD", "password")
@@ -113,6 +114,10 @@ AWS_CREDS = {
     "accessKeyId": os.environ.get("FLE_AWS_KEY", ""),
     "secretAccessKey": os.environ.get("FLE_AWS_SECRET", ""),
 }
+AWS_CREDS_2 = {
+    "accessKeyId": os.environ.get("FLE_AWS_KEY2", ""),
+    "secretAccessKey": os.environ.get("FLE_AWS_SECRET2", ""),
+}
 AZURE_CREDS = {
     "tenantId": os.environ.get("FLE_AZURE_TENANTID", ""),
     "clientId": os.environ.get("FLE_AZURE_CLIENTID", ""),
@@ -124,6 +129,9 @@ GCP_CREDS = {
 }
 KMIP_CREDS = {"endpoint": os.environ.get("FLE_KMIP_ENDPOINT", "localhost:5698")}
 
+# Ensure Evergreen metadata doesn't result in truncation
+os.environ.setdefault("MONGOB_LOG_MAX_DOCUMENT_LENGTH", "2000")
+
 
 def is_server_resolvable():
     """Returns True if 'server' is resolvable."""
@@ -281,6 +289,8 @@ class ClientContext:
     def client_options(self):
         """Return the MongoClient options for creating a duplicate client."""
         opts = client_context.default_client_options.copy()
+        opts["host"] = host
+        opts["port"] = port
         if client_context.auth_enabled:
             opts["username"] = db_user
             opts["password"] = db_pwd
@@ -378,7 +388,7 @@ class ClientContext:
                     self.auth_enabled = self._server_started_with_auth()
 
             if self.auth_enabled:
-                if not self.serverless:
+                if not self.serverless and not IS_SRV:
                     # See if db_user already exists.
                     if not self._check_user_provided():
                         _create_user(self.client.admin, db_user, db_pwd)
@@ -446,7 +456,7 @@ class ClientContext:
             else:
                 self.server_parameters = self.client.admin.command("getParameter", "*")
                 assert self.cmd_line is not None
-                if "enableTestCommands=1" in self.cmd_line["argv"]:
+                if self.server_parameters["enableTestCommands"]:
                     self.test_commands_enabled = True
                 elif "parsed" in self.cmd_line:
                     params = self.cmd_line["parsed"].get("setParameter", [])
@@ -482,14 +492,14 @@ class ClientContext:
 
     @property
     def host(self):
-        if self.is_rs:
+        if self.is_rs and not IS_SRV:
             primary = self.client.primary
             return str(primary[0]) if primary is not None else host
         return host
 
     @property
     def port(self):
-        if self.is_rs:
+        if self.is_rs and not IS_SRV:
             primary = self.client.primary
             return primary[1] if primary is not None else port
         return port
@@ -514,6 +524,10 @@ class ClientContext:
             # Raised if self.server_status is None.
             return None
 
+    def check_auth_type(self, auth_type):
+        auth_mechs = self.server_parameters.get("authenticationMechanisms", [])
+        return auth_type in auth_mechs
+
     def _check_user_provided(self):
         """Return True if db_user/db_password is already an admin user."""
         client: MongoClient = pymongo.MongoClient(
