@@ -87,7 +87,6 @@ class SrvPollingKnobs:
 
 
 class TestSrvPolling(unittest.TestCase):
-
     BASE_SRV_RESPONSE = [
         ("localhost.test.build.10gen.cc", 27017),
         ("localhost.test.build.10gen.cc", 27018),
@@ -96,8 +95,6 @@ class TestSrvPolling(unittest.TestCase):
     CONNECTION_STRING = "mongodb+srv://test1.test.build.10gen.cc"
 
     def setUp(self):
-        if not _HAVE_DNSPYTHON:
-            raise unittest.SkipTest("SRV polling tests require the dnspython module")
         # Patch timeouts to ensure short rescan SRV interval.
         self.client_knobs = client_knobs(
             heartbeat_frequency=WAIT_TIME,
@@ -112,7 +109,7 @@ class TestSrvPolling(unittest.TestCase):
     def get_nodelist(self, client):
         return client._topology.description.server_descriptions().keys()
 
-    def assert_nodelist_change(self, expected_nodelist, client):
+    def assert_nodelist_change(self, expected_nodelist, client, timeout=(100 * WAIT_TIME)):
         """Check if the client._topology eventually sees all nodes in the
         expected_nodelist.
         """
@@ -123,9 +120,9 @@ class TestSrvPolling(unittest.TestCase):
                 return True
             return False
 
-        wait_until(predicate, "see expected nodelist", timeout=100 * WAIT_TIME)
+        wait_until(predicate, "see expected nodelist", timeout=timeout)
 
-    def assert_nodelist_nochange(self, expected_nodelist, client):
+    def assert_nodelist_nochange(self, expected_nodelist, client, timeout=(100 * WAIT_TIME)):
         """Check if the client._topology ever deviates from seeing all nodes
         in the expected_nodelist. Consistency is checked after sleeping for
         (WAIT_TIME * 10) seconds. Also check that the resolver is called at
@@ -137,7 +134,7 @@ class TestSrvPolling(unittest.TestCase):
                 return pymongo.srv_resolver._SrvResolver.get_hosts_and_min_ttl.call_count >= 1
             return False
 
-        wait_until(predicate, "Node list equals expected nodelist", timeout=100 * WAIT_TIME)
+        wait_until(predicate, "Node list equals expected nodelist", timeout=timeout)
 
         nodelist = self.get_nodelist(client)
         if set(expected_nodelist) != set(nodelist):
@@ -151,6 +148,7 @@ class TestSrvPolling(unittest.TestCase):
         return True
 
     def run_scenario(self, dns_response, expect_change):
+        self.assertEqual(_HAVE_DNSPYTHON, True)
         if callable(dns_response):
             dns_resolver_response = dns_response
         else:
@@ -331,6 +329,28 @@ class TestSrvPolling(unittest.TestCase):
             with SrvPollingKnobs(nodelist_callback=nodelist_callback):
                 self.assert_nodelist_change(response, client)
 
+    def test_srv_waits_to_poll(self):
+        modified = [("localhost.test.build.10gen.cc", 27019)]
+
+        def resolver_response():
+            return modified
+
+        with SrvPollingKnobs(
+            ttl_time=WAIT_TIME,
+            min_srv_rescan_interval=WAIT_TIME,
+            nodelist_callback=resolver_response,
+        ):
+            client = MongoClient(self.CONNECTION_STRING)
+            self.assertRaises(
+                AssertionError, self.assert_nodelist_change, modified, client, timeout=WAIT_TIME / 2
+            )
+
+    def test_import_dns_resolver(self):
+        # Regression test for PYTHON-4407
+        import dns.resolver
+
+        self.assertTrue(hasattr(dns.resolver, "resolve"))
+
 
 if __name__ == "__main__":
     unittest.main()
